@@ -13,7 +13,7 @@ use 5.00800;
 use Carp 'croak';
 use Scalar::Util;
 
-our $VERSION = '0.21';
+our $VERSION = '0.22';
 our @ISA = qw(Exporter);
 our @EXPORT_OK = qw(encoded_string build_mt render_mt);
 our %EXPORT_TAGS = (
@@ -33,6 +33,7 @@ sub new {
         tag_start           => '<?',
         tag_end             => '?>',
         escape_func         => \&_inline_escape_html,
+        prepend             => '',
         package_name        => undef, # defaults to caller
         @_ == 1 ? ref($_[0]) ? %{$_[0]} : (template => $_[0]) : @_,
     }, $class;
@@ -136,6 +137,9 @@ sub _build {
             # Expression
             if ($type eq 'expr') {
                 my $escaped = $embed_escape_func->('$_MT_T');
+                if ($newline && $value =~ /\n/) {
+                    $value .= "\n"; # temporary workaround for t/13-heredoc.t
+                }
                 $lines[-1] .= "\$_MT_T = $value;\$_MT .= ref \$_MT_T eq 'Text::MicroTemplate::EncodedString' ? \$\$_MT_T : $escaped; \$_MT_T = '';";
             }
         }
@@ -176,8 +180,8 @@ sub parse {
 
     # Tokenize
     my $state = 'text';
-    my $push_expr = undef;
     my @lines = split /(\n)/, $tmpl;
+    my $tokens = [];
     while (@lines) {
         my $line = shift @lines;
         my $newline = undef;
@@ -210,24 +214,20 @@ sub parse {
         # Escaped line ending?
         if ($line =~ /(\\+)$/) {
             my $length = length $1;
-
             # Newline escaped
             if ($length == 1) {
                 $line =~ s/\\$//;
             }
-
             # Backslash escaped
             if ($length >= 2) {
                 $line =~ s/\\\\$/\\/;
                 $line .= "\n";
             }
+        } else {
+            $line .= "\n" if $newline;
         }
 
-        # Normal line ending
-        else { $line .= "\n" if $newline }
-
         # Mixed line
-        my @token;
         for my $token (split /
             (
                 $tag_start$expr_mark     # Expression
@@ -240,51 +240,43 @@ sub parse {
             )
         /x, $line) {
 
-            # Garbage
-            next if $token eq '';
-
-            # End
-            if ($token =~ /^$tag_end$/) {
+            # handle tags and bail out
+            if ($token eq '') {
+                next;
+            } elsif ($token =~ /^$tag_end$/) {
                 $state = 'text';
-                $push_expr = undef;
-            }
-
-            # Code
-            elsif ($token =~ /^$tag_start$/) { $state = 'code' }
-
-            # Comment
-            elsif ($token =~ /^$tag_start$cmnt_mark$/) { $state = 'cmnt' }
-
-            # Expression
-            elsif ($token =~ /^$tag_start$expr_mark$/) {
+                next;
+            } elsif ($token =~ /^$tag_start$/) {
+                $state = 'code';
+                next;
+            } elsif ($token =~ /^$tag_start$cmnt_mark$/) {
+                $state = 'cmnt';
+                next;
+            } elsif ($token =~ /^$tag_start$expr_mark$/) {
                 $state = 'expr';
+                next;
             }
 
-            # Value
-            else {
-
-                # Comments are ignored
-                next if $state eq 'cmnt';
-
-                if ($push_expr) {
-                    $push_expr->($token);
-                    next;
-                }
-
-                $state = 'code' if $push_expr;
-                if ($state eq 'expr') {
-                    my $token = \@token;
-                    $push_expr = sub {
-                        $token->[-1] .= $_[0];
-                    };
-                }
-
-                # Store value
-                push @token, $state, $token;
+            # value
+            if ($state eq 'text') {
+                push @$tokens, $state, $token;
+            } elsif ($state eq 'cmnt') {
+                next; # ignore comments
+            } elsif ($state eq 'cont') {
+                $tokens->[-1] .= $token;
+            } else {
+                # state is code or expr
+                push @$tokens, $state, $token;
+                $state = 'cont';
             }
         }
-        push @{$self->{tree}}, \@token;
+        if ($state eq 'text') {
+            push @{$self->{tree}}, $tokens;
+            $tokens = [];
+        }
     }
+    push @{$self->{tree}}, $tokens
+        if @$tokens;
     
     return $self;
 }
@@ -373,10 +365,12 @@ sub build {
         }
         '';
     }->();
+    my $line_offset = (() = ($_mt->{prepend} =~ /\n/sg)) + 5;
     my $expr = << "...";
 package $_mt->{package_name};
 sub {
-    ${_mt_setter}local \$SIG{__WARN__} = sub { print STDERR \$_mt->_error(shift, 4, \$_from) };
+    ${_mt_setter}local \$SIG{__WARN__} = sub { print STDERR \$_mt->_error(shift, $line_offset, \$_from) };
+    $_mt->{prepend}
     Text::MicroTemplate::encoded_string((
         $_code
     )->(\@_));
@@ -393,7 +387,7 @@ sub {
         if (my $_builder = eval($expr)) {
             return $_builder;
         }
-        $die_msg = $_mt->_error($@, 4, $_from);
+        $die_msg = $_mt->_error($@, $line_offset, $_from);
     }
     die $die_msg;
 }
@@ -582,6 +576,10 @@ escape function (defaults to L<Text::MicroTemplate::escape_html>), no escape whe
 
 package under where the renderer is compiled (defaults to caller package)
 
+=head3 prepend
+
+Prepends Perl code to the template.
+
 =head2 code()
 
 returns perl code that renders the template when evaluated
